#include "Meta.cuh"
#include <fstream>
void META::SetGrid(void) //
{
	if (usegrid)
	{
		std::vector<int> ngrid;
		std::vector<float> lower, upper;
		std::vector<bool> periodic;
		for (size_t i = 0; i < ndim; ++i)
		{
			ngrid.push_back(n_grids[i]);
			lower.push_back(cv_mins[i]);
			upper.push_back(cv_maxs[i]);
			bool is_periodic = cv_periods[i] > 0 ? true : false;
			periodic.push_back(is_periodic);
		}
		grid = new Grid<Gdata>(ngrid, lower, upper, periodic);
		potential_grid = new Grid<float>(ngrid, lower, upper, periodic);
		grid->data_ = vector<Gdata>(grid->size(), Gdata(grid->GetDimension(), 0.0));
		potential_grid->data_ = vector<float>(potential_grid->size(), 0.0);
	}
	else
	{
		printf("Warning! No grid version is very slow\n");
		grid = nullptr;
		potential_grid = nullptr;
	}
}
META::Hill::Hill(const Axis &centers, const Axis &inv_w, const Axis &period, const float &theight) : height(theight)
{
	for (int i = 0; i < centers.size(); ++i)
	{
		gsf.push_back(GaussianSF(centers[i], inv_w[i], period[i]));
	}
}
META::Gdata META::Hill::CalcHill(const Axis &values)
{
	const size_t &n = values.size();
	Axis dx(n, 0.0), df(n, 1.0);
	Gdata tder(n,1.0);

	// Compute difference between grid point and current val.
	for (size_t i = 0; i < n; ++i)
	{
		GaussianSF g = gsf[i];
		dx[i] = g.Evaluate(values[i], df[i]);
	}

	potential = 1.0;    // hill.potential
	// Compute derivative.
	for (size_t i = 0; i < n; ++i)
	{
		potential *= dx[i];
		for (size_t j = 0; j < n; ++j)
		{
			if (j != i)
			{
				tder[i] *= dx[j];
			}
			else
			{
				tder[i] *= df[j];
			}
		}
	}
	return tder;
}
void META::AddHill(const string &fn, Hill &hill)
{
	hills.push_back(hill);
	// Stop over printing when loading it!
	if (!fn.empty())
	{
		ofstream hillsout;
		hillsout.open(fn.c_str(), fstream::app);
		hillsout.precision(8);
		for (auto &gauss : hill.gsf)
		{
			hillsout << gauss.GetCenter() << "\t";
		}
		hillsout << hill.height << endl;
		hillsout.close();
	}
	// If grid is defined, add bias onto grid.
	if (usegrid) 
	{
		for (auto it = grid->begin(); it != grid->end(); ++it)
		{
			Gdata &data = *it;
			auto coord = it.coordinates();

			// Add to grid.
			Gdata tder = hill.CalcHill(coord);
			for (size_t i = 0; i < grid->GetDimension(); ++i)
			{
				data[i] += hill.height * tder[i];
			}
			potential_grid->at(coord) += hill.potential; 
		}
	}
}
void META::Estimate(const Axis &values, vector<Hill> &hills)
{
	if(usegrid)
	{
		potential_local = potential_grid->at(values);
		return;
	}
	potential_local = 0.;
	for (auto &hill : hills)
	{
		float tder = 1.;
		for (int i = 0; i < values.size(); ++i)
		{
			float temp; // df temp.
			tder *= hill.gsf[i].Evaluate(values[i], temp);
		}
		potential_local += tder * hill.height;
	}
}
void META::Initial(CONTROLLER *controller, COLLECTIVE_VARIABLE_CONTROLLER *cv_controller, char *module_name)
{
	if (module_name == NULL)
	{
		strcpy(this->module_name, "meta");
	}
	else
	{
		strcpy(this->module_name, module_name);
	}
	if (!cv_controller->Command_Exist(this->module_name, "CV"))
	{
		controller->printf("META1D IS NOT INITIALIZED\n\n");
		return;
	}
	else
	{
		std::vector<std::string> cv_str = cv_controller->Ask_For_String_Parameter(this->module_name, "CV", ndim);
		std::string cvv = std::accumulate(cv_str.begin(), cv_str.end(), std::string(""));
		printf("%s contains %d dimension META\n", cvv.c_str(), ndim);
	}
	cvs = cv_controller->Ask_For_CV(this->module_name, -1);
	if (cv_controller->Command_Exist(this->module_name, "Ndim"))
	{
		ndim = *cv_controller->Ask_For_Int_Parameter(this->module_name, "Ndim");
		if (ndim != cvs.size())
		{
			controller->printf("%d D-META IS NOT CONSISTANT CV size %d\n\n", ndim, cvs.size());
			return;
		}
	}
	else
	{
		ndim = cvs.size();
	}
	controller->printf("START INITIALIZING %dD-META:\n", ndim);
	// initialize Dpotential_local
	Malloc_Safely((void **)&Dpotential_local, sizeof(float) * ndim);
	sprintf(read_potential_file_name, "Meta_Potential.txt");
	sprintf(write_potential_file_name, "Meta_Potential.txt");
	cv_periods = cv_controller->Ask_For_Float_Parameter(this->module_name, "CV_period", cvs.size(), 1, false);
	cv_sigmas = cv_controller->Ask_For_Float_Parameter(this->module_name, "CV_sigma", cvs.size());
	cutoff = cv_controller->Ask_For_Float_Parameter(this->module_name, "CV_sigma", cvs.size());
	for (int i = 0; i < cvs.size(); i++)
	{
		if (cv_sigmas[i] <= 0)
		{
			controller->Throw_SPONGE_Error(spongeErrorValueErrorCommand, "METADYNAMICS::Initial", "CV_sigma should always be greater than 0");
		}
		cutoff[i] = 3 * cv_sigmas[i];
		cv_sigmas[i] = 1.0f / cv_sigmas[i]; /// cv_sigmas[i]; // inverted sigma!!!!!!!!!!!!!!!!!!!
	}
	if (cv_controller->Command_Exist(this->module_name, "nogrid"))
	{
		usegrid = false;
	}
	if (cv_controller->Command_Exist(this->module_name, "potential_in_file"))
	{
		strcpy(read_potential_file_name, cv_controller->Ask_For_String_Parameter(this->module_name, "potential_in_file")[0].c_str());
		Read_Potential(controller);
	}
	else
	{
		cv_mins = cv_controller->Ask_For_Float_Parameter(this->module_name, "CV_minimal", cvs.size());
		cv_maxs = cv_controller->Ask_For_Float_Parameter(this->module_name, "CV_maximum", cvs.size());
		n_grids = cv_controller->Ask_For_Int_Parameter(this->module_name, "CV_grid", cvs.size());
		total_grids = 1;
		for (int i = cvs.size() - 1; i > -1; i -= 1)
		{
			total_grids *= n_grids[i];
			if (cv_maxs[i] <= cv_mins[i])
			{
				controller->Throw_SPONGE_Error(spongeErrorValueErrorCommand, "METADYNAMICS::Initial", "CV_maximum should always be greater than CV_minimal");
			}
			if (n_grids[i] <= 1)
			{
				controller->Throw_SPONGE_Error(spongeErrorValueErrorCommand, "METADYNAMICS::Initial", "CV_grid should always be greater than 1");
			}
			float delta = (cv_maxs[i] - cv_mins[i]) / n_grids[i];
			cv_deltas.insert(cv_deltas.begin(), delta);
		}
	}
	/*Cuda_Malloc_And_Copy_Safely((void **)&d_cv_mins, cv_mins, sizeof(float) * cvs.size());
	Cuda_Malloc_And_Copy_Safely((void **)&d_cv_maxs, cv_maxs, sizeof(float) * cvs.size());
	Cuda_Malloc_And_Copy_Safely((void **)&d_cv_deltas, &cv_deltas[0], sizeof(float) * cvs.size());
	Cuda_Malloc_And_Copy_Safely((void **)&d_cv_periods, cv_periods, sizeof(float) * cvs.size());
	Cuda_Malloc_And_Copy_Safely((void **)&d_cv_sigmas, cv_sigmas, sizeof(float) * cvs.size());
	Cuda_Malloc_And_Copy_Safely((void **)&d_n_grid_suffix_product, &n_grid_suffix_product[0], sizeof(int) * cvs.size());
	
	*/
	void **temp_ptr = NULL;
	Malloc_Safely((void **)&temp_ptr, sizeof(void *) * cvs.size());
	for (int i = 0; i < cvs.size(); i += 1)
	{
		temp_ptr[i] = cvs[i]->d_value;
	}
	Cuda_Malloc_And_Copy_Safely((void **)&d_cv_values, temp_ptr, sizeof(void *) * cvs.size());
	for (int i = 0; i < cvs.size(); i += 1)
	{
		temp_ptr[i] = cvs[i]->crd_grads;
	}
	Cuda_Malloc_And_Copy_Safely((void **)&d_crd_grads, temp_ptr, sizeof(void *) * cvs.size());
	for (int i = 0; i < cvs.size(); i += 1)
	{
		temp_ptr[i] = cvs[i]->box_grads;
	}
	Cuda_Malloc_And_Copy_Safely((void **)&d_box_grads, temp_ptr, sizeof(void *) * cvs.size());
	free(temp_ptr);
	/*
		cv_period = 0;
		if (cv_controller->Command_Exist(this->module_name, "CV_period"))
		{
			float *temp_value = cv_controller->Ask_For_Float_Parameter(this->module_name, "CV_period");
			cv_period = temp_value[0];
			free(temp_value);
		}

		if (cv_controller->Command_Exist(this->module_name, "potential_in_file"))
		{
			strcpy(read_potential_file_name, cv_controller->Ask_For_String_Parameter(this->module_name, "potential_in_file")[0].c_str());
			Read_Potential(controller);
		}
		else
		{
			int check_necessary_inpu_exist = 0;
			if (cv_controller->Command_Exist(this->module_name, "CV_minimal"))
			{
				float *temp_value = cv_controller->Ask_For_Float_Parameter("meta", "CV_minimal");
				cv_min = temp_value[0];
				free(temp_value);
				check_necessary_inpu_exist += 1;
			}
			if (cv_controller->Command_Exist(this->module_name, "CV_maximum"))
			{
				float *temp_value = cv_controller->Ask_For_Float_Parameter("meta", "CV_maximum");
				cv_max = temp_value[0];
				free(temp_value);
				check_necessary_inpu_exist += 1;
			}
			dcv = 0.01;
			if (cv_controller->Command_Exist(this->module_name, "dCV"))
			{
				float *temp_value = cv_controller->Ask_For_Float_Parameter("meta", "dCV");
				dcv = temp_value[0];
				free(temp_value);
			}
			if ((float)(cv_max - cv_min) / dcv < 1.0)
			{
				check_necessary_inpu_exist = -1;
			}
			if (check_necessary_inpu_exist != 2)
			{
				char error_reason[CHAR_LENGTH_MAX];
				sprintf(error_reason, "Reason:\n\tthe required inputs are not complete or incorrect \
	(potential_in_file or (CV_minimal, CV_maximum))\n");
				controller->Throw_SPONGE_Error(spongeErrorMissingCommand, "META1D::Initial", error_reason);
			}

			grid_numbers = (int)(cv_max - cv_min/ dcv) + 1.;
			//dcv = (float)(cv_max - cv_min) / (grid_numbers - 1);

			Malloc_Safely((void **)&potential_list, sizeof(float) * grid_numbers);
			for (int i = 0; i < grid_numbers; i = i + 1)
			{
				potential_list[i] = 0;
			}
			printf("%lu", sizeof(potential_list));
			Malloc_Safely((void **)&factor_cv_grid_list, sizeof(float) * grid_numbers);
			for (int i = 0; i < grid_numbers; i = i + 1)
			{
				factor_cv_grid_list[i] = 0;
			}
		}
	*/
	if (cv_controller->Command_Exist(this->module_name, "height"))
	{
		float *temp_value = cv_controller->Ask_For_Float_Parameter("meta", "height");
		height_0 = temp_value[0];
		free(temp_value);
	}
	if (cv_controller->Command_Exist(this->module_name, "wall_height"))
	{
		float *temp_value = cv_controller->Ask_For_Float_Parameter("meta", "wall_height");
		border_potential_height = temp_value[0];
		free(temp_value);
	}
	if (cv_controller->Command_Exist(this->module_name, "welltemp_factor"))
	{
		float *temp_value = cv_controller->Ask_For_Float_Parameter("meta", "welltemp_factor");
		welltemp_factor = temp_value[0];
		free(temp_value);
		if (welltemp_factor > 1)
		{
			is_welltemp = 1;
		}
		else
		{
			printf("\tWell-tempered Factor must larger than 1!\n");
			getchar();
		}
	}
	if (cv_controller->Command_Exist(this->module_name, "potential_out_file"))
	{
		strcpy(write_potential_file_name, cv_controller->Ask_For_String_Parameter(this->module_name, "potential_out_file")[0].c_str());
	}
	if (cv_controller->Command_Exist(this->module_name, "potential_update_interval"))
	{
		int *temp_value = cv_controller->Ask_For_Int_Parameter(this->module_name, "potential_update_interval");
		potential_update_interval = temp_value[0];
		free(temp_value);
	}
	else
	{
		controller->printf("    Potential update interval is set to write_information_interval by default\n");
		if (controller->Command_Exist("write_information_interval"))
		{
			potential_update_interval = atoi(controller->Command("write_information_interval"));
		}
		else
		{
			potential_update_interval = 1000;
		}
	}

	controller->Step_Print_Initial("metaNd", "%.2f");
	controller->Step_Print_Initial("rbias", "%f");
	controller->Step_Print_Initial("rct", "%f");
	SetGrid();
	controller->printf("    potential output file: %s\n", write_potential_file_name);
	is_initialized = 1;
	controller->printf("END INITIALIZING META\n\n");
}

void META::Write_Potential()
{
	if (!is_initialized)
	{
		return;
	}
	FILE *temp_file = NULL;
	Open_File_Safely(&temp_file, write_potential_file_name, "w");
	fprintf(temp_file, "%dD-Meta\n", ndim);
	for (int i = 0; i < ndim; ++i)
	{
		fprintf(temp_file, "%f\t%f\t%f\n", cv_mins[i], cv_maxs[i], cv_deltas[i]);
	}
	vector<string> axis;
	int gridsize = 1;
	for (int i = 0; i < ndim; ++i)
	{
		fprintf(temp_file, " %d\t", n_grids[i]);
		gridsize *= n_grids[i];
	}
	fprintf(temp_file, "%d\n", gridsize);
	int index = 0;
	if (usegrid)
	{
		for (Grid<float>::iterator g_iter = potential_grid->begin(); g_iter != potential_grid->end(); ++g_iter)
		{
			auto values = g_iter.coordinates();
			ostringstream ss;
			for (const auto &v : values)
			{
				ss << v << "\t";
			}
			fprintf(temp_file, "%s%f\n", ss.str().c_str(), *g_iter);
		}
	}
	else
	{
		//Gdata data(grid->size(),0.0);
		vector<float> loop_flag(ndim, 0);
		for (int i = 0; i < ndim; ++i)
		{
			loop_flag[i] = cv_mins[i];
		}

		int i = 0;
		while (i >= 0)
		{
			Estimate(loop_flag, hills); //
			ostringstream ss;
			for (const float &v : loop_flag)
			{
				ss << v << "\t";
			}
			fprintf(temp_file, "%s%f\n", ss.str().c_str(), potential_local);
			//axis.push_back(ss.str());
			// iterate over any dimensions
			i = ndim - 1;
			while (i >= 0)
			{
				loop_flag[i] += cv_deltas[i];
				if (loop_flag[i] > cv_maxs[i] + 0.0000001)
				{
					loop_flag[i] = cv_mins[i];
					--i;
				}
				else
				{
					break;
				}
			}
		}
	}
	fclose(temp_file);
}
void META::Read_Potential(CONTROLLER *controller)
{
	FILE *temp_file = NULL;
	Open_File_Safely(&temp_file, read_potential_file_name, "r");
	char temp_char[256];
	char *get_val = fgets(temp_char, 256, temp_file);
	for (int i = 0; i < ndim; ++i)
	{
		int scanf_ret1 = fscanf(temp_file, "%f %f %f\n", &cv_mins[i], &cv_maxs[i], &cv_deltas[i]);
		if (scanf_ret1 != 3)
		{
			controller->Throw_SPONGE_Error(spongeErrorBadFileFormat, "META::Read_Potential", "Reason:\n\tbad potential input file\n");
		}
		controller->printf("    CV_minimal = %f\n    CV_maximum = %f\n    dCV = %f\n", cv_mins[i], cv_maxs[i], cv_deltas[i]);
	}
	for (int i = 0; i < ndim; ++i)
	{
		int scanf_ret2 = fscanf(temp_file, "%d\n", &n_grids[i]);
		if (scanf_ret2 != 1)
		{
			controller->Throw_SPONGE_Error(spongeErrorBadFileFormat, "META::Read_Potential", "Reason:\n\tbad potential input file\n");
		}
	}
	/*
	Malloc_Safely((void **)&potential_list, sizeof(float) * grid_numbers);
	Malloc_Safely((void **)&factor_cv_grid_list, sizeof(float) * grid_numbers);
	for (int i = 0; i < grid_numbers; i = i + 1)
	{
		factor_cv_grid_list[i] = 0;
	}
	float temp_float;
	for (int i = 0; i < grid_numbers; i = i + 1)
	{
		int scanf_ret3 = fscanf(temp_file, "%f %f\n", &temp_float, &potential_list[i]);
		if (scanf_ret3 != 2)
		{
			controller->Throw_SPONGE_Error(spongeErrorBadFileFormat, "META::Read_Potential", "Reason:\n\tbad potential input file\n");
		}
	}
	*/
	fclose(temp_file);
}

static __global__ void Add_Frc(const int atom_numbers, VECTOR *frc, VECTOR *cv_grad, float dheight_dcv)
{
	for (int i = blockIdx.x + blockDim.x * threadIdx.x; i < atom_numbers; i += gridDim.x * blockDim.x)
	{
		frc[i] = frc[i] - dheight_dcv * cv_grad[i];
		// printf("The Force: dheight_dcv %f*{%f,%f,%f} of %d atom\n",dheight_dcv,cv_grad[i].x,cv_grad[i].y,cv_grad[i].z,i);
	}
}

static __global__ void Add_Potential(float *d_potential, const float to_add)
{
	d_potential[0] += to_add;
}

static __global__ void Add_Virial(float *d_virial, const float dU_dCV, const VECTOR *dCV_dL, const VECTOR box_length)
{
	d_virial[0] -= dU_dCV * dCV_dL[0] * box_length;
}

void META::Meta_Force_With_Energy_And_Virial(int atom_numbers, VECTOR *frc, int need_potential, int need_pressure, float *d_potential, float *d_virial, VECTOR box_length)
{
	if (!is_initialized)
	{
		return;
	}
	Potential_and_derivative(need_potential);

	for (int i = 0; i < cvs.size(); ++i)
	{
		Add_Frc<<<1, 1>>>(atom_numbers, frc, cvs[i]->crd_grads, Dpotential_local[i]);
		if (need_pressure)
		{
			Add_Virial<<<1, 1>>>(d_virial, Dpotential_local[i], cvs[i]->box_grads, box_length);
		}
	}
	if (need_potential)
	{
		Add_Potential<<<1, 1>>>(d_potential, potential_local); // device to host add!!!
	}
}

void META::Potential_and_derivative(const int need_potential)
{
	if (!is_initialized)
	{
		return;
	}
	Axis values; 
	for (int i = 0; i < cvs.size(); ++i)
	{
		values.push_back(cvs[i]->value);
		Dpotential_local[i] = 0.f;
	}
	if (need_potential)
	{
		Estimate(values, hills);
	}
	if (usegrid)
	{
		for (int i = 0; i < cvs.size(); ++i)
		{
			float gvi =  grid->at(values)[i];
			Dpotential_local[i] = gvi;
		}
	}
	else
	{
		for (auto &hill : hills)
		{
			for (int i = 0; i < cvs.size(); ++i)
			{
				Dpotential_local[i] += hill.height * hill.CalcHill(values)[i];
			}
		}
	}
}

float exp_added(float expsum, const float expvalue)
{
	if (expsum > expvalue)
	{
		expsum = expsum + logf(1.0 + expf(expvalue - expsum));
	}
	else
	{
		expsum = expvalue + logf(1.0 + expf(expsum - expvalue));
	}
	return expsum;
}

void META::Border_derivative(float *border_upper, float *border_lower, float *cutoff, float *Dpotential_local)
{
	for (int i = 0; i < cvs.size(); ++i)
	{
		float h_cv = cvs[i]->value;
		if (h_cv - border_lower[i] < cutoff[i])
		{
			Dpotential_local[i] = Dpotential_local[i] - border_potential_height * expf(border_lower[i] - h_cv);
		}
		else if (border_upper[i] - h_cv < cutoff[i])
		{
			Dpotential_local[i] = Dpotential_local[i] + border_potential_height * expf(h_cv - border_upper[i]);
		}
	}
}

void META::computeReweightingFactor(float temp)
{
	if (welltemp_factor == 1.0)
	{ // in this case we have no bias, so reweight factor is 0.0
		rct = 0.;
	}

	float Z_0 = 0.; // proportional to the integral of exp(-beta*F)
	float Z_V = 0.; // proportional to the integral of exp(-beta*(F+V))
	float minusBetaFplusV = 1. / (welltemp_factor - 1.) / CONSTANT_kB / temp;
	float minusBetaF = welltemp_factor * minusBetaFplusV;
	if (welltemp_factor == -1.0)
	{ // non well-tempered case
		minusBetaF = 1;
		minusBetaFplusV = 0;
	}
        if(usegrid)
        {
		for (Grid<float>::iterator g_iter = potential_grid->begin(); g_iter != potential_grid->end(); ++g_iter)
		{
			Z_0 = exp_added(Z_0, *g_iter * minusBetaF);
			Z_V = exp_added(Z_V, *g_iter * minusBetaFplusV);
		}
        }
        else
	{
		vector<float> loop_flag(ndim, 0);
		for (int i = 0; i < ndim; ++i)
		{
			loop_flag[i] = cv_mins[i];
		}

		int i = 0;
		while (i >= 0)
		{
			Estimate(loop_flag, hills); //
			Z_0 = exp_added(Z_0, potential_local * minusBetaF);
			Z_V = exp_added(Z_V, potential_local * minusBetaFplusV);
			// iterate over any dimensions
			i = ndim - 1;
			while (i >= 0)
			{
				loop_flag[i] += cv_deltas[i];
				if (loop_flag[i] > cv_maxs[i] + 0.0000001)
				{
					loop_flag[i] = cv_mins[i];
					--i;
				}
				else
				{
					break;
				}
			}
		}
	}
	rct = CONSTANT_kB * temp * (Z_0 - Z_V);
}

void META::getReweightingBias(float temp)
{
	computeReweightingFactor(temp);
	rbias =  potential_local - rct;
}

void META::getHeight(float temp)
{
	height = height_0;
	if (temp < 0.00001)
	{
		return; // avoid /0 = nan
	}
	if (is_welltemp == 1)
	{
		float bias_old = potential_local;
		float height_welltemp = height_0 * expf(-bias_old / (welltemp_factor - 1) / CONSTANT_kB / temp);
		height = height_welltemp;
	}
}

void META::AddPotential(float temp, int steps)
{
	if (!is_initialized)
	{
		return;
	}
	if (steps % potential_update_interval == 0)
	{
		Axis values, sigmas, periods;
		for (int i = 0; i < cvs.size(); ++i)
		{
			values.push_back(cvs[i]->value);
			sigmas.push_back(cv_sigmas[i]);
			periods.push_back(cv_periods[i]);
		}
		Estimate(values,hills);
		getHeight(temp);
		getReweightingBias(temp);
		Hill newhill = Hill(values, sigmas, periods, height);
		AddHill("", newhill);
	}
}

void META::Do_Metadynamics(int atom_numbers, UNSIGNED_INT_VECTOR *uint_crd, VECTOR scaler, VECTOR *crd, VECTOR box_length,
			   int step, int need_potential, int need_pressure, VECTOR *frc, float *d_potential, float *d_virial, float sys_temp)
{
	if (this->is_initialized)
	{
		int need = CV_NEED_CPU_VALUE | CV_NEED_CRD_GRADS;
		if (need_pressure)
		{
			need |= CV_NEED_BOX_GRADS;
		}

		for (int i = 0; i < cvs.size(); i = i + 1)
		{
			this->cvs[i]->Compute(atom_numbers, uint_crd, scaler, crd, box_length, need, step);
		}
		this->Meta_Force_With_Energy_And_Virial(atom_numbers, frc, need_potential, need_pressure, d_potential, d_virial, box_length);
		AddPotential(sys_temp, step);
	}
}
